/*
 * hdw-sniff, sniffer using pcap lib
 *
 * author: hackbard@hackdaworld.dyndns.org
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pcap.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/ioctl.h>

/* IEEE 802.3 stuff -- i will concentrate on .11 stuff before! */
#include <netinet/if_ether.h>  /* for ethhdr struct */
#include <netinet/ip.h> /* ip */
#include <netinet/in.h>	/* in_addr , inet_ntoa */

/* IEEE 802.11 stuff -- will become one include later ... */
#include "ieee80211.h" /* from hunz's aeolus, short hostap_wlan.h */
#include "ieee802_11.h" /* from pcmcia-cs */

#include "hdw-sniff.h" /* my functions */

int main(int argc, char *argv[]) {

	char pcap_error[PCAP_ERRBUF_SIZE];
	pcap_t *pcap_handle;
	int pcap_fd,foo_fd;
	fd_set pcap_fd_set;
	struct timeval fd_set_tv;
	char sys_call[30];
	FILE *logfile;
	struct info_struct my_info_struct;
	
	/* parse the arguments */
	if(argc<3) {
		printf("usage: %s <interface> <monitor mode> <logfile>\n",
			argv[0]);
		return 0;
	}
	if(argc!=4) {
		printf("no logfile specified, writing to stdout ...\n");
	} 
	else {
		if((logfile=fopen(argv[3],"w"))!=NULL) {
			printf("writing to logfile %s ...\n",argv[3]);
		}
		else {
			printf("can't open logfile!\n");
		}
	}
			
	/* setting up device and set monitor mode */
	if(atoi(argv[2])==1) {
		printf("setting to monitor mode\n");
		if(strncmp(argv[1],"wlan",4)==0)
			sprintf(sys_call,"iwpriv %s monitor 3",argv[1]);
		if(strncmp(argv[1],"eth",3)==0)
			sprintf(sys_call,"ifconfig %s promisc",argv[1]);
	system(sys_call);
	}
	printf("setting up interface\n");
	sprintf(sys_call,"ifconfig %s up",argv[1]);
	system(sys_call);

	/* start pcap session */
	pcap_handle=pcap_open_live(argv[1],BUFSIZ,1,-1,pcap_error);
	if(pcap_handle==NULL) {
		printf("%s: %s\n",argv[0],pcap_error);
		return 1;
	}

	/* set non blocking */
	if((pcap_setnonblock(pcap_handle,1,pcap_error))==-1) {
		printf("%s: %s\n",argv[0],pcap_error);
		return 1;
	}
	
	/* grab a package until user breaks */
	my_info_struct.count=0;
	my_info_struct.mmode=argv[2][0];
	strcpy(my_info_struct.dev,argv[1]);

	/* prepare for select */
	pcap_fd=pcap_fileno(pcap_handle);

	/* create file descriptor */
	if((foo_fd=socket(AF_INET,SOCK_DGRAM,0))==-1) {
		printf("unable to create socket foo_fd\n");
		return -1;
	}

	/* do loopp */
	while (1) {

		/* set pcap_fd */
		FD_ZERO(&pcap_fd_set);
		FD_SET(pcap_fd,&pcap_fd_set);
		fd_set_tv.tv_sec=0;
		fd_set_tv.tv_usec=200000;

		if((select(pcap_fd+1,&pcap_fd_set,NULL,NULL,&fd_set_tv)) && (FD_ISSET(pcap_fd,&pcap_fd_set))) pcap_dispatch(pcap_handle,-1,pcap_process,(u_char *)&my_info_struct);
		else if((hop_channel(&my_info_struct,foo_fd))==-1) { 
			printf("channelhopping failed, aborting\n");
			return -1;
		}
	}
	return 0;
}


/* pcap_process callback function */
void pcap_process(u_char *info,const struct pcap_pkthdr *pcap_header,
			const u_char *package) {
	
	/* local variables */
	char tmp_buf[20];
	struct linux_wlan_ng_prism_hdr *prism_hdr;
	struct ieee802_11_hdr *w_hdr;
	struct ethhdr *e_hdr;
	struct iphdr *ip_hdr;
	struct info_struct *my_info_struct;
	int i,p_o,w_o,e_o,i_o;
	
	my_info_struct=(struct info_struct *)info;
	++(my_info_struct->count);
	
	/* cache offsets */
	p_o=((strncmp(my_info_struct->dev,"wlan",4)==0)?sizeof(struct linux_wlan_ng_prism_hdr):0);
	w_o=((strncmp(my_info_struct->dev,"wlan",4)==0)?sizeof(struct ieee802_11_hdr):0);
	e_o=sizeof(struct ethhdr);
	i_o=sizeof(struct iphdr);

	/* new package */
	printf("\n");
	printf("---> package %d ---- %s",my_info_struct->count,
			ctime((const time_t*)&(pcap_header->ts.tv_sec)));
	
	/* pcap header */
	printf("pcap header: ");
	printf("capture_length(dec): %d\t",pcap_header->caplen);
	printf("off_wire_length(dec): %d\n",pcap_header->len);

	/* wireless stuff */
	/* prism wlan ng headers */
	if((my_info_struct->mmode-0x30==1) && 
		(strncmp(my_info_struct->dev,"wlan",4)==0)) {
	printf("prism header: (%d bytes)\n",p_o);
	prism_hdr=(struct linux_wlan_ng_prism_hdr *)package;
	printf("| message code/length: %d/%d |\t",prism_hdr->msgcode,
						prism_hdr->msglen);
	printf("device: %s |\n",prism_hdr->devname);
	/* ieee802.11 header */
	printf("ieee802.11 header: (%d bytes)\n",w_o);
	w_hdr=(struct ieee802_11_hdr *)(package+p_o);
	
	printf("fc: ");
	for(i=0;i<16;i++)
		printf("%s%d%s",(i==0?"|":""),
				(((w_hdr->frame_ctl) & (1<<i))>0?1:0),
				(i==15?"|\n":"|"));
	printf("    | v | t |  s-t  |t|f|m|r|p|m|w|o|\n");
	/* frame type */
	/* management */
	if(!(w_hdr->frame_ctl & 0x0c)) {
	if((w_hdr->frame_ctl & IEEE802_11_STYPE_ASSOC_REQ)>0)
		strcpy(tmp_buf,"association request");
	else if((w_hdr->frame_ctl & IEEE802_11_STYPE_ASSOC_RESP)>0)
		strcpy(tmp_buf,"association response");
	else if((w_hdr->frame_ctl & IEEE802_11_STYPE_REASSOC_REQ)>0)
		strcpy(tmp_buf,"reassociation request");
	else if((w_hdr->frame_ctl & IEEE802_11_STYPE_REASSOC_RESP)>0)
		strcpy(tmp_buf,"reassociation response");
	else if((w_hdr->frame_ctl & IEEE802_11_STYPE_PROBE_REQ)>0)
		strcpy(tmp_buf,"probe request");
	else if((w_hdr->frame_ctl & IEEE802_11_STYPE_PROBE_RESP)>0)
		strcpy(tmp_buf,"probe response");
	else if((w_hdr->frame_ctl & IEEE802_11_STYPE_BEACON)>0)
		strcpy(tmp_buf,"beacon");
	else if((w_hdr->frame_ctl & IEEE802_11_STYPE_ATIM)>0)
		strcpy(tmp_buf,"announcement traffic indication message");
	else if((w_hdr->frame_ctl & IEEE802_11_STYPE_DISASSOC)>0)
		strcpy(tmp_buf,"disassociation");
	else if((w_hdr->frame_ctl & IEEE802_11_STYPE_AUTH)>0)
		strcpy(tmp_buf,"authentification");
	else if((w_hdr->frame_ctl & IEEE802_11_STYPE_DEAUTH)>0)
		strcpy(tmp_buf,"deauthentification");
	else strcpy(tmp_buf,"impossible situation \%) - go mail the author.");
	}
	else strcpy(tmp_buf,"control & monitor frame types not supported yet");
	/* print out frame type */
	printf("=> %s\n",tmp_buf);
	
	printf("duration/id: 0x%x\n",w_hdr->duration_id);
	printf("version check ... %s\n",
	((w_hdr->frame_ctl & IEEE802_11_FCTL_VERS)==0x00)?
							"ok":"unknown");
	}

	/* ieee802.3 */
	/* ethernet */
	if((strncmp(my_info_struct->dev,"eth",3)==0) | 
		((w_hdr->frame_ctl & IEEE802_11_FTYPE_DATA)>0)) {
	printf("ethernet: (%d bytes)\n",e_o);
	e_hdr=(struct ethhdr *)(package+p_o+w_o);
	/* what types ? */
	printf("type = ");
	printf("%x  ",ntohs(e_hdr->h_proto));
	printf("dest_addr = ");
	for(i=0;i<ETH_ALEN;i++)
	printf("%x%s",*(e_hdr->h_dest+i),((i==ETH_ALEN-1)?" ":":"));
	printf(" src_addr = ");
	for(i=0;i<ETH_ALEN;i++)
	printf("%x%s",*(e_hdr->h_source+i),((i==ETH_ALEN-1)?"\n":":"));

		/* IP ? */
	if(ntohs(e_hdr->h_proto)==ETH_P_IP) {
		printf("ip protocol: (%d bytes)\n",i_o);
		ip_hdr=(struct iphdr *)(package+p_o+w_o+e_o);
		printf("version = %x ",ntohs(ip_hdr->version));
		printf("header_length = %x \n",ntohs(ip_hdr->ihl));
		printf("service = %x ",ntohs(ip_hdr->tos));
		printf("total_length(dec.) = %d \n",ntohs(ip_hdr->tot_len));
		printf("source_ip: ");
		for(i=0;i<=3;++i) {
			printf("%d%s",
			(ip_hdr->saddr&(0xff<<(8*i)))>>(8*i),
			(i==3?"\n":"."));
		}
		printf("destination_ip: ");
		for(i=0;i<=3;++i) {
			printf("%d%s",
			(ip_hdr->daddr&(0xff<<(8*i)))>>(8*i),
			(i==3?"\n":"."));
		}
		printf("ip_id = %x ",ntohs(ip_hdr->id));
		printf("ip_offset = %x \n",ntohs(ip_hdr->frag_off));
		printf("time2live = %x ip_proto = %x\n",ntohs(ip_hdr->ttl),
						ntohs(ip_hdr->protocol));
		// printf("chksum: %x\n",ntohs(ip_hdr->ip_sum));
	}
	}

	/* check what we have ... */
	printf("all dump: (hex)\n");
	for(i=p_o+w_o;i<pcap_header->caplen;i++)
		printf("%x ",*(package+i));
	printf("\n");
	printf("all dump: (char)\n");
	for(i=p_o+w_o;i<pcap_header->caplen;i++)
		printf("%c ",*(package+i));
	printf("\n");
}

int hop_channel(struct info_struct *info,int foo_fd) {
	if((info->mmode-0x30==1) && (strncmp(info->dev,"wlan",4)==0)) {

	struct iwreq my_iwreq;

	if (info->channel>=14) info->channel=1;

	memset(&my_iwreq,0,sizeof(my_iwreq));
	strcpy(my_iwreq.ifr_name,info->dev);
	printf("debug: channel = %d\n",info->channel);
	my_iwreq.u.freq.e=0;
	my_iwreq.u.freq.m=info->channel;
	if((ioctl(foo_fd,SIOCSIWFREQ,&my_iwreq))==-1) {
		printf("unable to hop channels\n");
		perror("ioctl");
		return -1;
	}
	++(info->channel);
	}
	return 0;
}
